// as_scripting.
// Written by Dominque Douglas
// Copyright 2013 Dominque Douglas
#include "DXUT.h"
#include "as_scripting.h"
#include "scriptd3dxcolor.h"
#include "scriptd3dxmathfunc.h"
#include "scriptgamedialoginterface.h"
#include "scriptcamerainterface.h"
#include "scriptgamestate.h"
#include "scriptdroidammostates.h"
#include "scriptaudiointerface.h"

// this will allow our scripts to be able to use std::string
#include "scriptstdstring/scriptstdstring.h"
#include "scriptarray/scriptarray.h"


// Implement a simple message callback function
void MessageCallback(const asSMessageInfo *msg, void *param)
{
	const char *type = "ERR ";
	if( msg->type == asMSGTYPE_WARNING ) type = "WARN";
	else if( msg->type == asMSGTYPE_INFORMATION ) type = "INFO";

	char real_message[256];

	sprintf_s(real_message, 256, "%s (%d, %d) : %s : %s\n", msg->section, msg->row, msg->col, type, msg->message);

	// this is a windows application so use a message box
	MessageBoxA(NULL, real_message, "AngelScript Message", MB_OK);
}

// Print the script string to the standard output stream
// For Testing
void print(std::string &msg)
{
	MessageBoxA(NULL, msg.c_str(), "AngelScript Message", MB_OK);
}

int StartScriptingSystem(asIScriptEngine *&scriptengine, ScriptContextData &contextdata)
{
	int result;

	// Create the script engine
	scriptengine = asCreateScriptEngine(ANGELSCRIPT_VERSION);

	// Set the message callback to receive information on errors in human readable form.
	result = scriptengine->SetMessageCallback(asFUNCTION(MessageCallback), 0, asCALL_CDECL);
	if(result < 0) return result; // an error has occurred

	result = RegisterGameInterface(scriptengine);
	if(result < 0) return result; // an error has occurred

	result =  LoadScript(scriptengine, contextdata);
	if(result < 0) return result; // an error has occurred

	contextdata.ctx = scriptengine->CreateContext();

	return result;
}

void ShutdownScriptingSystem(asIScriptEngine *&scriptengine, asIScriptContext *&ctx)
{
	// Why check to see if this is NULL? This function will be called at the end of the program
	// as a way to clean it up even if an error has occurred. If an error occurs during initialization
	// one or both of these variables may be null.
	if(ctx)
	{
		ctx->Release();
		ctx = NULL; // i don't like leaving old pointers that don't point to valid data
	}

	if(scriptengine)
	{
		scriptengine->Release();
		scriptengine = NULL; // i don't like leaving old pointers that don't point to valid data
	}
}

int RegisterGameInterface(asIScriptEngine *scriptengine)
{
	int result;

	// Set Acces Mask ---------------------------------------------------------------------
	scriptengine->SetDefaultAccessMask(ScriptInterfaceMask_All); 
	// Bindings made in this section are accessible to everything -------------------------

	// Register STD String (Needed to help me test my implementation)
	RegisterStdString(scriptengine);

	// Register Arrays
	RegisterScriptArray(scriptengine, true);

	// Register print function (Needed to help me test my implementation)
	result = scriptengine->RegisterGlobalFunction("void print(const string &in)", asFUNCTION(print), asCALL_CDECL);
	if(result < 0) return result;

	// Set Acces Mask ---------------------------------------------------------------------
	scriptengine->SetDefaultAccessMask(ScriptInterfaceMask_Gameplay); 
	// Bindings made in this section are accessible to game play related scripts ----------

	// Register enum GAME_MODE
	result = RegisterEnumGAME_MODE(scriptengine);
	if(result < 0) return result;

	result = RegisterD3DXMathFunctions(scriptengine);
	if(result < 0) return result;

	result = RegisterD3DXCOLOR(scriptengine);
	if(result < 0) return result;

	result = RegisterDROID_STATE(scriptengine);
	if(result < 0) return result;

	result = RegisterAMMO_STATE(scriptengine);
	if(result < 0) return result;

	result = RegisterGameStateInterface(scriptengine);
	if(result < 0) return result;

	result = RegisterAudioInterface(scriptengine);
	if(result < 0) return result;

	result = RegisterCameraInterface(scriptengine); // FireAmmo() needs to get the view
	if(result < 0) return result;

	// Set Acces Mask ---------------------------------------------------------------------
	scriptengine->SetDefaultAccessMask(ScriptInterfaceMask_SetupOnly); 
	// Bindings made in this section are accessible to setup related scripts --------------

	result = RegisterDialogInterface(scriptengine);

	return result;
}

int RegisterEnumGAME_MODE(asIScriptEngine *scriptengine)
{
	int result;

	result = scriptengine->RegisterEnum("GAME_MODE");
	if(result < 0) return result;

	result = scriptengine->RegisterEnumValue("GAME_MODE", "GAME_RUNNING", (int)GAME_RUNNING);
	if(result < 0) return result;

    result = scriptengine->RegisterEnumValue("GAME_MODE", "GAME_MAIN_MENU", (int)GAME_MAIN_MENU);
	if(result < 0) return result;

    result = scriptengine->RegisterEnumValue("GAME_MODE", "GAME_AUDIO_MENU", (int)GAME_AUDIO_MENU);
	if(result < 0) return result;

    result = scriptengine->RegisterEnumValue("GAME_MODE", "GAME_VIDEO_MENU", (int)GAME_VIDEO_MENU);

	return result;
}

asIScriptFunction *GetScriptFunction(asIScriptModule *mod, char *declaration)
{
	asIScriptFunction *func = mod->GetFunctionByDecl(declaration);
	if(func == NULL)
	{
		MessageBoxA(NULL,"AngelScript Message - The script function missing. Please add it and try again.", declaration, MB_OK);
		return NULL;
	}
	return func;
}

int LoadScript(asIScriptEngine *scriptengine, ScriptContextData &contextdata)
{
	int result;

	// The CScriptBuilder helper is an add-on that loads the file,
	// performs a pre-processing pass if necessary, and then tells
	// the engine to build a script module.
	CScriptBuilder builder;

	// load initapp.as into a module -----------------------------------------------------------------------------
	result = builder.StartNewModule(scriptengine, "InitAppModule"); 
	if( result < 0 ) 
	{
		// If the code fails here it is usually because there
		// is no more memory to allocate the module
		MessageBoxA(NULL, "Unrecoverable error while starting a new module.", "AngelScript Message", MB_OK);
		return result;
	}

	// set the modules access mask
	builder.GetModule()->SetAccessMask(ScriptInterfaceMask_All); // give init app full access

	// load the script
	result = builder.AddSectionFromFile("initapp.as");
	if( result < 0 ) 
	{
		// The builder wasn't able to load the file. Maybe the file
		// has been removed, or the wrong name was given, or some
		// preprocessing commands are incorrectly written.
		MessageBoxA(NULL,"Please correct the errors in the script and try again.", "AngelScript Message", MB_OK);
		return result;
	}
	result = builder.BuildModule();
	if( result < 0 ) 
	{
		// An error occurred. Instruct the script writer to fix the 
		// compilation errors that were listed in the output stream.
		MessageBoxA(NULL,"Please correct the errors in the script and try again.", "AngelScript Message", MB_OK);
		return result;
	}

	// load gamelogic.as into a module ----------------------------------------------------------------------
	result = builder.StartNewModule(scriptengine, "GameModule"); 
	if( result < 0 ) 
	{
		// If the code fails here it is usually because there
		// is no more memory to allocate the module
		MessageBoxA(NULL, "Unrecoverable error while starting a new module.", "AngelScript Message", MB_OK);
		return result;
	}

	// set the modules access mask
	builder.GetModule()->SetAccessMask(ScriptInterfaceMask_Gameplay); // only give gameplay access

	// load the script
	result = builder.AddSectionFromFile("gamelogic.as");
	if( result < 0 ) 
	{
		// The builder wasn't able to load the file. Maybe the file
		// has been removed, or the wrong name was given, or some
		// preprocessing commands are incorrectly written.
		MessageBoxA(NULL,"Please correct the errors in the script and try again.", "AngelScript Message", MB_OK);
		return result;
	}
	result = builder.BuildModule();
	if( result < 0 ) 
	{
		// An error occurred. Instruct the script writer to fix the 
		// compilation errors that were listed in the output stream.
		MessageBoxA(NULL,"Please correct the errors in the script and try again.", "AngelScript Message", MB_OK);
		return result;
	}

	// Find the function that is to be called. 
	asIScriptModule *modInitApp = scriptengine->GetModule("InitAppModule");
	contextdata.script_functions[Function_InitApp] = GetScriptFunction(modInitApp, "void InitApp()");
	if( contextdata.script_functions[Function_InitApp] == 0 ) return -1;

	asIScriptModule *modlogic = scriptengine->GetModule("GameModule");
	contextdata.script_functions[Function_FireAmmo] = GetScriptFunction(modlogic, "void FireAmmo()");
	if( contextdata.script_functions[Function_FireAmmo] == 0 ) return -1;

	contextdata.script_functions[Function_HandleAmmoAI] = GetScriptFunction(modlogic, "void HandleAmmoAI( float fElapsedTime )");
	if( contextdata.script_functions[Function_HandleAmmoAI] == 0 ) return -1;

	contextdata.script_functions[Function_HandleDroidAI] = GetScriptFunction(modlogic, "void HandleDroidAI( float fElapsedTime )");
	if( contextdata.script_functions[Function_HandleDroidAI] == 0 ) return -1;

	contextdata.script_functions[Function_CreateDroid] = GetScriptFunction(modlogic, "void CreateDroid()");
	if( contextdata.script_functions[Function_CreateDroid] == 0 ) return -1;

	// load the constants
	contextdata.constants.SetContantsFromScript(modlogic);

	return result;
}


